/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2017  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QFile>
#include <QFileInfo>
#include <QDateTime>
#include <QDir>
#include <assert.h>

#include "../../core/xmlfile.h"

#include "filedata.h"

TFileDataNode::TFileDataNode(
	TFileDataNode 						*parent)
{
	if(parent)
	{
		d_is_directory = false;
		parent->Add_Child(this);
	}
}

TFileDataNode::~TFileDataNode(void)
{

}

int TFileDataNode::Children_Count(void) const
{
	return static_cast<int>(d_children.size());
}

TFileDataNode* TFileDataNode::Child(
	const int 							offset) const
{
	assert(static_cast<unsigned int>(offset) < d_children.size());

	return d_children[offset];
}

void TFileDataNode::Set_Directory(
	const bool 						state)
{
	if(!state)
	{
		this->Clear();
	}

	d_is_directory = state;
}

void TFileDataNode::Add_Child(
	TFileDataNode						*node)
{
	assert(node);

	if(d_is_directory)
	{
		d_children.push_back(node);
	}
}

void TFileDataNode::Set_File_Name(
	const QString 						&file_name)
{
	d_file_name = file_name;
}

void TFileDataNode::Clear(void)
{
	std::vector<TFileDataNode*>::iterator	iter;

	if(d_is_directory)
	{
		for(iter = d_children.begin();iter < d_children.end();++iter)
		{
			(*iter)->Clear();
			delete (*iter);
		}

		d_children.clear();
	}
}

TFileData::TFileData(void)
{

}

TFileData::~TFileData(void)
{
	d_root_node.Clear();
}

bool TFileData::Save_File_List(
	const QString						&source_path,
	const QString 						&file_list_name) const
{
	QDomElement							node;
	QDomElement							node_lev_a;
	QFileInfo							file_info;
	QDateTime							file_date_time;
    QStringList::const_iterator			iter;
	TXmlFile							xmlfile;
	
	try
	{
		xmlfile.SetFileName(file_list_name);
		xmlfile.Reset("File_Sync_Data",1);
		
		file_date_time.setTimeSpec(Qt::UTC);
		
		node = xmlfile.CreateNode(xmlfile.RootNode(),"File_List");
		
		for(iter = d_source_file_names.constBegin();iter < d_source_file_names.constEnd();++iter)
		{
			file_info.setFile(source_path + '/' + (*iter));
			
			if(file_info.exists())
			{
				node_lev_a = xmlfile.CreateNode(&node,"File");
				
				file_date_time = file_info.lastModified();

				xmlfile.WriteAttribute(&node_lev_a,"Modified",static_cast<int>(file_date_time.toTime_t()));
				xmlfile.WriteTextNode(&node_lev_a,(*iter));
			}
		}
		
		xmlfile.WriteXmlFile();
	}
	catch(TXmlFileException exception)
	{
		d_last_error = exception.ErrorText();
		return false;
	}

	return true;
}

int TFileData::Update_File_Count(void) const
{
	return d_update_file_names.size();
}

QString TFileData::Get_Update_File_Name(
	const QString 						&path,
	const int 							offset) const
{
	assert(offset < d_update_file_names.size());
	assert(!(offset < 0));

	return path + d_update_file_names[offset];
}

int TFileData::Source_File_Count(void) const
{
	return d_source_file_names.size();
}

QString TFileData::Get_Source_File_Name(
	const QString 						&path,
	const int 							offset) const
{
	assert(offset < d_source_file_names.size());
	assert(!(offset < 0));

	return path + '/' + d_source_file_names[offset];
}

bool TFileData::Load_File_List(
	const QString 						&path,
	const QString 						&file_list_name)
{
	TXmlFile							xmlfile;
	int									version;

	d_source_file_hash.clear();
	
	try
	{
		xmlfile.SetFileName(path + '/' + file_list_name);
		xmlfile.Open("File_Sync_Data",&version);
		
		switch(version)
		{
			case 1:
				this->Load_File_List_V1(&xmlfile);
				break;
				
			default:
				d_last_error = "File_Sync_Data file version not supported";
				return false;
		}

	}
	catch(TXmlFileException exception)
	{
		d_last_error = exception.ErrorText();
		return false;
	}
	
   	return true;
}

bool TFileData::Create_File_List(
	const QString 						&path)
{
	QFileInfo							file_info(path);

	if(!file_info.exists())
	{
		d_last_error = QString("Source path '%1' does not exist").arg(file_info.absoluteFilePath());
		return false;
	}

	if(!file_info.isDir())
	{
		d_last_error = QString("Source path '%1' must be a directory").arg(file_info.absoluteFilePath());
		return false;
	}

	d_root_node.Clear();
	d_root_node.Set_File_Name(QString());

	d_root_node.Set_Directory(true);
	this->Create_File_List(&d_root_node,path);

	this->Update_Source_File_Names();

	return true;
}

bool TFileData::Create_Update_File_List(
	const QString 						&path)
{
	QHash<QString,unsigned int>::const_iterator	iter;
	unsigned int						U1,U2;
	QFileInfo							file_info;
	QDateTime							file_date_time;

	if(d_source_file_hash.empty())
	{
		d_last_error = QString("Source list is empty");
		return false;
	}

	d_update_file_names.clear();
	file_date_time.setTimeSpec(Qt::UTC);

	for(iter = d_source_file_hash.begin();iter != d_source_file_hash.end();++iter)
	{
		file_info.setFile(path + '/' + iter.key());

		if(file_info.exists())
		{
			U1 = iter.value();
			file_date_time = file_info.lastModified();
			U2 = file_date_time.toTime_t();
			
			if((U1-U2) > 1 && ( (U1-U2) > 3601 || (U1-U2) < 3599) )	// one hour variation for daylight savings and one second variation might be seen between different computers for some reason
			{
				d_update_file_names.push_back(iter.key());
			}
		}
		else
		{
			d_update_file_names.push_back(iter.key());
		}
	}

	return true;
}

void TFileData::Update_Source_File_Names(void)
{
	d_source_file_names.clear();

	this->Update_Source_File_Names(QString(),&d_source_file_names,&d_root_node);
}

void TFileData::Create_File_List(
	TFileDataNode 						*parent,
	const QString 						&path)
{
	QDir								dir_path(path);
	QFileInfo							file_info;
	QStringList							file_list;
	QStringList::const_iterator			iter;
	TFileDataNode						*node;

	assert(parent);

	file_list = dir_path.entryList(QDir::NoDotAndDotDot | QDir::Dirs | QDir::Files | QDir::Hidden);

	for(iter = file_list.begin();iter < file_list.end();++iter)
	{
		if((*iter) != QStringLiteral(".DS_Store"))	// do not include this file
		{
			file_info.setFile(dir_path.absolutePath() + '/' + (*iter));
			
			node = new TFileDataNode(parent);

			node->Set_File_Name(*iter);

			if(file_info.isDir())
			{
				node->Set_Directory(true);
				this->Create_File_List(node,file_info.absolutePath() + '/' + node->File_Name());
			}
			else
			{
				node->Set_Directory(false);
			}
		}
	}
}

void TFileData::Update_Source_File_Names(
	const QString						&prefix,
	QStringList 						*list,
	TFileDataNode						*node)
{
	std::vector<TFileDataNode*>::const_iterator iter;
	QString										name;

	assert(list);
	assert(node);

	if(prefix.length())
	{
		name.append(prefix + '/' + node->File_Name());
	}
	else
	{
		name.append(node->File_Name());
	}

	if(!node->Is_Directory())
	{
		list->push_back(name);
		return;
	}

	for(iter = node->d_children.begin();iter < node->d_children.end();++iter)
	{
		this->Update_Source_File_Names(name,list,(*iter));
	}
}

void TFileData::Load_File_List_V1(
	TXmlFile							*xmlfile)
{
	QDomElement							node;
	QDomElement							node_lev_a;
	QString								modified;
	QString								file_name;

	assert(xmlfile);
	
	node = xmlfile->GetNode(xmlfile->RootNode(),"File_List");
	node_lev_a = xmlfile->GetNode(&node,"File",true);

	while(!node_lev_a.isNull())
	{
		file_name= xmlfile->ReadTextNode(&node_lev_a);
		modified = xmlfile->ReadAttribute(&node_lev_a,"Modified");
		
		d_source_file_hash[file_name] = modified.toUInt();
		
		node_lev_a = xmlfile->GetSiblingNode(&node_lev_a,"File",true);
	}
}

